
#include <m_pd.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <inttypes.h>
#include <fcntl.h>

#if defined __WIN32__
# include <windows.h>
#endif

#if defined __APPLE__ || defined __linux__
# include <termios.h>
#endif

#define EP_KEY_LENGTH 32

#define BOARDS_MAX 8

#define RECEIVERS_MAX 512

#define TTY_NAME_LENGTH 128

static t_class *Arduino_class;

typedef struct {
	t_object x_obj;

	enum { ep_invalid, ep_none, ep_send, ep_receive } endpoint_type;

	union {
		struct {
			t_clock *x_clock;

			char tty[TTY_NAME_LENGTH];
			long baudrate;
#if defined __WIN32__
			HANDLE hSerial;
#endif
#if defined __APPLE__ || defined __linux__
			int fd;
#endif
			char connected;

			char inputBufferTag[EP_KEY_LENGTH];
			float inputBufferValue;

			uint8_t inputBufferState;
			uint8_t inputBufferTagIdx;
			uint8_t *inputBufferValuePtr;
			uint8_t inputBufferValuePtrIdx;
		} b;

		struct {
			t_outlet *f_out;

			char endpoint_tag[EP_KEY_LENGTH];
		} r;

		struct {
			t_outlet *f_out;

			char endpoint_tag[EP_KEY_LENGTH];
		} s;
	} d;
} t_Arduino;

static struct {
	t_Arduino *boards[BOARDS_MAX];

	t_Arduino *receivers[RECEIVERS_MAX];

} Arduino;

void Arduino_tick(t_Arduino *x) {
	// default only go here once a second
	int delay = 1000;

	// need to re-connect or to receive data?
	if(!x->d.b.connected) {
		do {
			post("Arduino_tick %p connecting %s ...", x, x->d.b.tty);

			if(x->d.b.baudrate < 1200 || x->d.b.baudrate > 230400) {
				post("Arduino_tick %p WARNING: baudrate %ld might be out of range!\n", x->d.b.baudrate);
			}

#if defined __WIN32__
			// open serial port
			x->d.b.hSerial = CreateFile(x->d.b.tty, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
			if(x->d.b.hSerial == INVALID_HANDLE_VALUE) {
				post("Arduino_tick %p CreateFile(%s) failed", x, x->d.b.tty);
				break;
			}

			// setup comm configure
			DCB dcb;
			dcb.DCBlength = sizeof(dcb);

			// fetch current comm settings
			if(!GetCommState(x->d.b.hSerial, &dcb)) {
				post("Arduino_tick %p GetCommState(%s) failed", x, x->d.b.tty);
				CloseHandle(x->d.b.hSerial);
				break;
			}

			// modify comm settings  TODO: obtain baud setting from patch
			dcb.BaudRate = x->d.b.baudrate;
			dcb.ByteSize = 8;
			dcb.StopBits = ONESTOPBIT;
			dcb.Parity = NOPARITY;

			// set back comm settings
			if(!SetCommState(x->d.b.hSerial, &dcb)) {
				post("Arduino_tick %p SetCommState(%s) failed", x, x->d.b.tty);
				CloseHandle(x->d.b.hSerial);
				break;
			}

			COMMTIMEOUTS CommTimeouts;

			// fetch current timeout settings
			if(!GetCommTimeouts(x->d.b.hSerial, &CommTimeouts)) {
				post("Arduino_tick %p GetCommTimeouts(%s) failed", x, x->d.b.tty);
				CloseHandle(x->d.b.hSerial);
				break;
			}

			// modify timeout settings  TODO: obtain baud setting from patch
			CommTimeouts.ReadIntervalTimeout = 1;
			CommTimeouts.ReadTotalTimeoutConstant = 1;
			CommTimeouts.ReadTotalTimeoutMultiplier = 1;
			CommTimeouts.WriteTotalTimeoutConstant = 5000;
			CommTimeouts.WriteTotalTimeoutMultiplier = 1000;

			// set back timeout settings
			if(!SetCommTimeouts(x->d.b.hSerial, &CommTimeouts)) {
				post("Arduino_tick %p SetCommTimeouts(%s) failed", x, x->d.b.tty);
				CloseHandle(x->d.b.hSerial);
				break;
			}
#endif

#if defined __APPLE__ || defined __linux__
      // open device
      x->d.b.fd = open(x->d.b.tty, O_RDWR | O_NOCTTY | O_NONBLOCK);
			if(x->d.b.fd < 0) {
				post("Arduino_tick %p open(%s) failed", x, x->d.b.tty);
				break;
			}

			struct termios newtio;

			// get current value
			int r = tcgetattr(x->d.b.fd, &newtio);
			if(r < 0) { 
				post("Arduino_tick %p tcgetattr(%s) failed", x, x->d.b.tty);
				close(x->d.b.fd);
				break;
			}

			// set baudrate
#ifdef __APPLE__
			newtio.c_cflag = CS8 | CLOCAL | CREAD;
			newtio.c_ispeed = x->d.b.baudrate;
			newtio.c_ospeed = x->d.b.baudrate;
#else
			newtio.c_cflag = CS8 | CLOCAL | CREAD;
			newtio.c_ispeed = B115200; // FIXME: look-up
			newtio.c_ospeed = B115200; // FIXME: look-up
#endif
			newtio.c_iflag = IGNPAR;
			newtio.c_oflag = 0;
			newtio.c_lflag = 0; // ICANON;
			newtio.c_cc[VMIN] = 1;
			newtio.c_cc[VTIME] = 0;

#ifdef __APPLE__
			cfsetospeed(&newtio, x->d.b.baudrate);
			cfsetispeed(&newtio, x->d.b.baudrate);
#else
			cfsetospeed(&newtio, B115200); // FIXME: look-up
			cfsetispeed(&newtio, B115200); // FIXME: look-up
#endif

			// flush data
			r = tcflush(x->d.b.fd, TCIFLUSH);
			if(r < 0) { 
				post("Arduino_tick %p tcflush(%s, TCIFLUSH) failed", x, x->d.b.tty);
				close(x->d.b.fd);
				break;
			}

			// set new value
			r = tcsetattr(x->d.b.fd, TCSADRAIN, &newtio);
			if(r < 0) { 
				post("Arduino_tick %p tcsetattr(%s, TCSADRAIN) failed", x, x->d.b.tty);
				close(x->d.b.fd);
				break;
			}

      // set non-blocking
      r = fcntl(x->d.b.fd, F_GETFL);
			if(r < 0) { 
				post("Arduino_tick %p fcntl(%s, F_GETFL) failed", x, x->d.b.tty);
				close(x->d.b.fd);
				break;
			}
			r = fcntl(x->d.b.fd, F_SETFL, r | O_NONBLOCK | FNDELAY);
			if(r < 0) { 
				post("Arduino_tick %p fcntl(%s, F_SETFL) failed", x, x->d.b.tty);
				close(x->d.b.fd);
				break;
			}
#endif

			// connection succeeded
			x->d.b.connected = 1;

			// reset state
			x->d.b.inputBufferState = 0;

			post("Arduino_tick %p connected %s ok!", x, x->d.b.tty);

			// only read data every 100ms
			delay = 100;
		} while(0);
	} else {
		for(;;) {
			uint8_t inputData = 0;

#if defined __WIN32__
			// check for incomming messages
			long nread = 0;
			if(!ReadFile(x->d.b.hSerial, &inputData, sizeof(inputData), &nread, NULL)) {
				post("Arduino_tick %p ReadFile(%s) failed, closing connection", x, x->d.b.tty);
				CloseHandle(x->d.b.hSerial);
				x->d.b.connected = 0;
				break;
			}

			if(!nread) break;
#endif

#if defined __APPLE__ || defined __linux__
      // try to read data
      ssize_t nread = read(x->d.b.fd, &inputData, sizeof(inputData));

      // failed?
      if(nread < 0) {
        // don't see error on no data
        if(errno == EAGAIN) break;

				// show error
				post("Arduino_tick %p read(%s) failed '%s', closing connection", x, x->d.b.tty, strerror(errno));
				close(x->d.b.fd);
				x->d.b.connected = 0;
				break;
			}
#endif

			switch(x->d.b.inputBufferState) {
				// wait for start byte
				case(0):
				{
					if(inputData != 0xAA) break;

					x->d.b.inputBufferTagIdx = 0;
					x->d.b.inputBufferValuePtrIdx = 0;
					x->d.b.inputBufferValuePtr = (uint8_t *)&x->d.b.inputBufferValue;

					x->d.b.inputBufferState++;
					break;
				}
				// wait for reception of tag to complete
				case(1):
				{
					if(inputData != 0x00) {
						// always keep room for a null-terminator
						if(x->d.b.inputBufferTagIdx < (sizeof(x->d.b.inputBufferTag)-1)) {
							x->d.b.inputBufferTag[x->d.b.inputBufferTagIdx++] = inputData;
						}
						break;
					}

					// close string
					x->d.b.inputBufferTag[x->d.b.inputBufferTagIdx++] = '\0';
					x->d.b.inputBufferState++;
					break;
				}
				// wait for the floating-point value to complete
				case(2):
				{
					if(x->d.b.inputBufferValuePtrIdx < sizeof(x->d.b.inputBufferValue)) {
						x->d.b.inputBufferValuePtr[x->d.b.inputBufferValuePtrIdx++] = inputData;
					}

					if(x->d.b.inputBufferValuePtrIdx < sizeof(x->d.b.inputBufferValue)) break;

					// find a matching receiver
					int i;
					for(i = 0; i < RECEIVERS_MAX; i++) {
						// receiver configured?
						t_Arduino * const y = Arduino.receivers[i];
						if(y == NULL) continue;

						// receiver matches tag?
						if(strcmp(x->d.b.inputBufferTag, y->d.r.endpoint_tag) != 0) continue;

						// output the value to this matching receiver
						const t_float f = x->d.b.inputBufferValue;
						outlet_float(y->d.r.f_out, f);
					}

					x->d.b.inputBufferState = 0;
					break;
				}
			}
		}

		// only read data every 100ms
		delay = 100;
	}

	// re-schedule
	clock_delay(x->d.b.x_clock, delay);
}

void *Arduino_new(t_symbol *s, int argc, t_atom *argv) {
	post("Arduino_new start");

	t_Arduino *x = (t_Arduino *)pd_new(Arduino_class);
	if(x == NULL) return(NULL);

	t_symbol *cmd = (argc)?atom_getsymbol(argv+0):NULL;

	if(argc >= 2 && strcmp(cmd->s_name, "open") == 0) {
		post("Arduino_new 'open' %p", x);

		// add board to the list
		int i;
		for(i = 0; i < BOARDS_MAX; i++) {
			t_Arduino * const y = Arduino.boards[i];
			if(y == NULL) break;
		}

		// any room for this board at all?
		if(i >= BOARDS_MAX) {
			// TODO: destroy this board
			return(NULL);
		}

		// store this board
		Arduino.boards[i] = x;

		// this object is not used to send or receive data
		x->endpoint_type = ep_none;

		// get tty to connect
		t_symbol *sym = atom_getsymbol(argv+1);
		strncpy(x->d.b.tty, sym->s_name, sizeof(x->d.b.tty));

		// get baudrate to run at
		x->d.b.baudrate = (long)((argc >= 3)?atom_getfloat(argv+2):115200L);

		// have a clock for managing the connection
		x->d.b.x_clock = clock_new(&x->x_obj, (t_method)Arduino_tick);
		clock_delay(x->d.b.x_clock, 0);

		// default we are not connected
		x->d.b.connected = 0;

	} else if(argc == 2 && strcmp(cmd->s_name, "send") == 0) {
		post("Arduino_new 'send' %p", x);

		// store our endpoint type and tag
		x->endpoint_type = ep_send;
		t_symbol *sym = atom_getsymbol(argv+1);
		strncpy(x->d.s.endpoint_tag, sym->s_name, sizeof(x->d.s.endpoint_tag));

		// also output the value we send for eas of use
	  x->d.s.f_out = outlet_new(&x->x_obj, &s_float);

	} else if(argc == 2 && strcmp(cmd->s_name, "receive") == 0) {
		post("Arduino_new 'receive' %p", x);

		// store our endpoint type and tag
		x->endpoint_type = ep_receive;
		t_symbol *sym = atom_getsymbol(argv+1);
		strncpy(x->d.r.endpoint_tag, sym->s_name, sizeof(x->d.r.endpoint_tag));

		// create an outlet to write floats on
	  x->d.r.f_out = outlet_new(&x->x_obj, &s_float);

		// add receiver to the list
		int i;
		for(i = 0; i < RECEIVERS_MAX; i++) {
			t_Arduino * const y = Arduino.receivers[i];
			if(y == NULL) break;
		}

		// any room for this receiver at all?
		if(i >= RECEIVERS_MAX) {
			// TODO: destroy this receiver
			return(NULL);
		}

		// store this receiver
		Arduino.receivers[i] = x;

	} else {
		// report error but still create the object so they can request help
		post("error: Arduino: unknown or missing arguments: <open|send|receive> [...]");
		
		// mark endpoint invalid
		x->endpoint_type = ep_invalid;
	}

	post("Arduino_new end");

	return((void *)x);
}

static uint8_t Arduino_send_raw(t_Arduino * const y, const uint8_t * const data, const long length, long * const written) {
#if defined __WIN32__
	return(WriteFile(y->d.b.hSerial, data, length, written, NULL));
#endif
#if defined __APPLE__ || defined __linux__
	return(write(y->d.b.fd, data, length) == length);
#endif
}

void Arduino_send_float(t_Arduino *x, t_float value) {
	if(x->endpoint_type != ep_send) {
		post("error: Arduino: no method for 'float'");
		return;
	}

	post("Arduino_send %f to %s", value, x->d.s.endpoint_tag);

	int i;
	for(i = 0; i < BOARDS_MAX; i++) {
		t_Arduino * const y = Arduino.boards[i];
		if(y == NULL) continue;

		long written = 0;

		const uint8_t start = 0xAA;
		if(!Arduino_send_raw(y, &start, sizeof(start), &written)) {
			post("Arduino_send_float WriteFile(%s, start) failed", y->d.b.tty);
			continue;
		}

		if(!Arduino_send_raw(y, (const uint8_t * const)x->d.s.endpoint_tag, strlen(x->d.s.endpoint_tag)+1, &written)) {
			post("Arduino_send_float WriteFile(%s, tag) failed", y->d.b.tty);
			continue;
		}

		const float _v = value;
		if(!Arduino_send_raw(y, (const uint8_t * const)&_v, sizeof(_v), &written)) {
			post("Arduino_send_float WriteFile(%s, value) failed", y->d.b.tty);
			continue;
		}
	}

	// also output the value we send for eas of use
	outlet_float(x->d.s.f_out, value);
}

void Arduino_bang(t_Arduino *x) {
	post("Arduino_bang!");

	if(x->endpoint_type != ep_none) {
		post("error: Arduino: no method for 'bang'");
		return;
	}
}

void Arduino_free(t_Arduino *x) {
	post("Arduino_free %p", x);

	// figure out what is been freed
	int i;
	for(i = 0; i < BOARDS_MAX; i++) {
		t_Arduino * const y = Arduino.boards[i];
		if(y != x) continue;

		// prevent from being scheduled again
		clock_free(y->d.b.x_clock);

#if defined __WIN32__
		post("Arduino_free CloseHandle(%s)", y->d.b.tty);
		CloseHandle(y->d.b.hSerial);
#endif

#if defined __APPLE__ || defined __linux__
		post("Arduino_free close(%s)", y->d.b.tty);
		close(x->d.b.fd);
#endif

		post("Arduino_free Arduino.boards[%d] = NULL", i);
		Arduino.boards[i] = NULL;
	}
	for(i = 0; i < RECEIVERS_MAX; i++) {
		t_Arduino * const y = Arduino.receivers[i];
		if(y != x) continue;

		post("Arduino_free Arduino.receivers[%d] = NULL", i);
		Arduino.receivers[i] = NULL;
	}
}

void Arduino_setup(void) {
	post("Arduino_setup start");

	memset(&Arduino, 0, sizeof(Arduino));

	Arduino_class = class_new(gensym("Arduino"), (t_newmethod)Arduino_new, (t_method)Arduino_free, sizeof(t_Arduino), CLASS_DEFAULT, A_GIMME, 0);

  class_addbang(Arduino_class, Arduino_bang);

	class_addmethod(Arduino_class, (t_method)Arduino_send_float, gensym("float"), A_FLOAT, 0);

	class_sethelpsymbol(Arduino_class, gensym("help-Arduino"));

	post("Arduino_setup end");
}

